/* $Id: log.c,v 1.1.1.1 2005/05/28 17:24:26 bedlam Exp $ */
/* log.c
 * - Functions for logging in ices
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2001 Brendan Cully
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#include <time.h>
#include "definitions.h"

extern stream404_config_t stream404_config;

/* Private function declarations */
static void stream404_log_string (char *format, char *string);
static int stream404_log_open_logfile (void);
static int stream404_log_close_logfile (void);

static char lasterror[BUFSIZE];
/* Public function definitions */

/* Initialize the log module, creates log file */
void
stream404_log_initialize (void)
{
	if (!stream404_log_open_logfile ()) {
		stream404_log ("%s", stream404_log_get_error ());
	}
	stream404_log ("Logfile opened");
}

/* Shutdown the log module, close the logfile */
void
stream404_log_shutdown (void)
{
	if (!stream404_log_close_logfile ()) {
		stream404_log ("%s", stream404_log_get_error ());
	}
}

/* Close everything, start up with clean slate when 
 * run as a daemon */
void
stream404_log_daemonize (void)
{
	close (0); 
	close (1); 
	close (2);

	freopen ("/dev/null", "r", stdin);
	freopen ("/dev/null", "w", stdout);
	freopen ("/dev/null", "w", stderr);

	stream404_log_reopen_logfile ();
}

/* Cycle the logfile, usually called from the SIGHUP handler */
int
stream404_log_reopen_logfile (void)
{
	stream404_log_close_logfile ();
	return stream404_log_open_logfile ();
}

/* Log only if verbose mode is set. Prepend output with DEBUG:  */
void
stream404_log_debug (const char *fmt, ...)
{
	char buff[BUFSIZE];
	va_list ap;
	time_t now;
	char tmp[1024];
	char *timestamp;

	if (!stream404_config.verbose)
		return;

	va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(buff, BUFSIZE, fmt, ap);
#else
	vsprintf(buff, fmt, ap);
#endif
	va_end(ap);

	(void)time(&now);
	timestamp = ctime(&now) + 4;
	*(timestamp + 15) = '\0';
	snprintf(tmp, 1024, "%s DEBUG", timestamp);
	strncat(tmp, ": %s\n", 1024);
	stream404_log_string (tmp, buff);
}

/* Log to console and file */
void
stream404_log (const char *fmt, ...)
{
	char buff[BUFSIZE];
	va_list ap;
	time_t now;
	char tmp[1024];
	char *timestamp;

	va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(buff, BUFSIZE, fmt, ap);
#else
	vsprintf(buff, fmt, ap);
#endif
	va_end(ap);

	(void)time(&now);
	timestamp = ctime(&now) + 4;
	*(timestamp + 15) = '\0';
	snprintf(tmp, 1024, "%s", timestamp);
	strncat(tmp, " %s\n", 1024);
	stream404_log_string (tmp, buff);
}

/* Store error information in module memory */
void
stream404_log_error (const char *fmt, ...)
{
	char buff[BUFSIZE];
	va_list ap;
	time_t now;
	char tmp[BUFSIZE+16];
	char *timestamp;

	va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(buff, BUFSIZE, fmt, ap);
#else
	vsprintf(buff, fmt, ap);
#endif
	va_end(ap);

	(void)time(&now);
	timestamp = ctime(&now) + 4;
	*(timestamp + 15) = '\0';
	snprintf(tmp, 1024, "%s", timestamp);
	strncat(tmp, buff, BUFSIZE);
	strncpy (lasterror, tmp, BUFSIZE+16);
}

/* Get last error from log module */
char *
stream404_log_get_error (void)
{
	return lasterror;
}

/* Private function definitions */

/* Function to log string to both console and file */
static void
stream404_log_string (char *format, char *string)
{
	if (stream404_config.logfile) {
		fprintf (stream404_config.logfile, format, string);
#ifndef HAVE_SETLINEBUF
		fflush (stream404_config.logfile);
#endif
	}
	
	/* Don't log to console when daemonized */
	if (!stream404_config.daemon) {
		fprintf (stdout, format, string);
	}
}

/* Open the stream404 logfile, create it if needed */
static int
stream404_log_open_logfile (void)
{
  char namespace[1024], buf[1024];
  FILE *logfp;

  if (! stream404_config.base_directory ||
      strlen (stream404_config.base_directory) > 1016)
  {
    stream404_log_error ("Base directory is invalid");
    return 0;
  }

  snprintf (namespace, sizeof (namespace), "%s/stream404.log",
	    stream404_config.base_directory);

  logfp = fopen (namespace, "a");

  if (!logfp) {
    stream404_log_error ("Error while opening %s, error: %s", namespace,
		    stream404_util_strerror (errno, buf, 1024));
    return 0;
  }

  stream404_config.logfile = logfp;
#ifdef HAVE_SETLINEBUF
  setlinebuf (stream404_config.logfile);
#endif

  return 1;
}

/* Close stream404' logfile */
static int
stream404_log_close_logfile (void)
{
	if (stream404_config.logfile) {
		stream404_util_fclose (stream404_config.logfile);
	}

	return 1;
}
