/* $Id: pm_builtin.c,v 1.3 2005/07/18 18:14:58 bedlam Exp $ */
/* playlist_text.c
 * - Module for simple playlist
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2001 Brendan Cully
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <definitions.h>
#include "rand.h"

static FILE* fp_main = NULL; 
static FILE* fp_backup = NULL;
static time_t playlist_modtime_main = 0; 
static time_t playlist_modtime_backup = 0;
static int lineno_main = 0;
static int lineno_backup = 0;

extern stream404_config_t stream404_config;

/* Private function declarations */
static char* playlist_builtin_main_get_next(void *);
static char* playlist_builtin_backup_get_next(void *);
static int playlist_builtin_main_get_lineno (void);
static int playlist_builtin_backup_get_lineno (void);
static void playlist_builtin_main_shutdown(void);
static void playlist_builtin_backup_shutdown(void);

static int playlist_builtin_main_open_playlist (playlist_module_t* pm);
static int playlist_builtin_backup_open_playlist (playlist_module_t* pm);
static int playlist_builtin_main_reopen_playlist (playlist_module_t* pm);
static int playlist_builtin_backup_reopen_playlist (playlist_module_t* pm);
static void playlist_builtin_main_shuffle_playlist (void);
static void playlist_builtin_backup_shuffle_playlist (void);
static int playlist_builtin_line_skip (int lineno, FILE* fp);

#define MAIN 0
#define BACKUP 1

/* Global function definitions */

/* Initialize the builting playlist handler */
int
stream404_playlist_builtin_initialize (playlist_module_t* pm, int status)
{
    if (status == MAIN) {
	stream404_log_debug ("Initializing builtin playlist handler (%s)...", pm->playlist_file);
	
	pm->get_next = playlist_builtin_main_get_next;
	pm->get_metadata = NULL;
	pm->get_lineno = playlist_builtin_main_get_lineno;
	pm->shutdown = playlist_builtin_main_shutdown;
	
	if (!playlist_builtin_main_open_playlist (pm)) {
	    stream404_log ("Could not find a valid playlist file.");
	    stream404_setup_shutdown ();
	    return -1;
	}
	
	if (pm->randomize) {
	    stream404_log_debug ("Randomizing playlist...");
	    playlist_builtin_main_shuffle_playlist ();
	}
	
	lineno_main = 0;
	return 1;
    }
    else {
	stream404_log_debug ("Initializing builtin backup playlist handler (%s)...", pm->playlist_file);
	
	pm->get_next = playlist_builtin_backup_get_next;
	pm->get_metadata = NULL;
	pm->get_lineno = playlist_builtin_backup_get_lineno;
	pm->shutdown = playlist_builtin_backup_shutdown;
	
	if (!playlist_builtin_backup_open_playlist (pm)) {
	    stream404_log ("Could not find a valid backup playlist file.");
	    stream404_setup_shutdown ();
	    return -1;
	}
	
	if (pm->randomize) {
	    stream404_log_debug ("Randomizing backup playlist...");
	    playlist_builtin_backup_shuffle_playlist ();
	}
	
	lineno_backup = 0;
	return 1;
    }
}

/*
int
stream404_playlist_builtin_backup_initialize (playlist_module_t* pm)
{
  stream404_log_debug ("Initializing builtin backup playlist handler (%s)...", pm->playlist_file);

  pm->get_next = playlist_builtin_backup_get_next;
  pm->get_metadata = NULL;
  pm->get_lineno = playlist_builtin_backup_get_lineno;
  pm->shutdown = playlist_builtin_backup_shutdown;

  if (!playlist_builtin_backup_open_playlist (pm)) {
    stream404_log ("Could not find a valid backup playlist file.");
    stream404_setup_shutdown ();
    return -1;
  }

  if (pm->randomize) {
    stream404_log_debug ("Randomizing backup playlist...");
    playlist_builtin_backup_shuffle_playlist ();
  }

  lineno_backup = 0;
  return 1;
}
*/

static char *
playlist_builtin_main_get_next (void *arg)
{
  char *out;
  static int level = 0;
  struct stat st;
  input_stream_t *source = arg;

  /* If the original playlist has changed on disk (and we are not
   * randomising) reload it. */
  if ((!fp_main || (!stream404_config.pm.randomize &&
	       (stat (stream404_config.pm.playlist_file, &st) == 0) &&
	       (st.st_mtime > playlist_modtime_main))) &&
      !playlist_builtin_main_reopen_playlist (&stream404_config.pm))
    return NULL;

  if (feof (fp_main)) {
    stream404_log_debug ("Reached end of playlist, rewinding");
    lineno_main = 0;
    rewind (fp_main);
  }

  if (! (out = stream404_util_read_line (fp_main)))
    return NULL;

  if (out[0])
    out[strlen (out) - 1] = '\0';

  if (! out[0]) {
    if (level++) {
      level = 0;
      stream404_log_error ("Unreadable or empty playlist");
      return NULL;
    }
    return playlist_builtin_main_get_next (arg);
  }

  level = 0;

  lineno_main++;

  stream404_log_debug ("Builtin playlist handler serving: %s", stream404_util_nullcheck (out));

  if (strncasecmp(out, "http://", 7) == 0)
    source->location = REMOTE;

  return out;
}

static char *
playlist_builtin_backup_get_next (void *arg)
{
  char *out;
  static int level = 0;
  struct stat st;
  input_stream_t *source = arg;

  /* If the original playlist has changed on disk (and we are not
   * randomising) reload it. */
  if ((!fp_backup || (!stream404_config.pm.randomize &&
	       (stat (stream404_config.pm.playlist_file, &st) == 0) &&
	       (st.st_mtime > playlist_modtime_backup))) &&
      !playlist_builtin_backup_reopen_playlist (&stream404_config.pm))
    return NULL;

  if (feof (fp_backup)) {
    stream404_log_debug ("Reached end of playlist, rewinding");
    lineno_backup = 0;
    rewind (fp_backup);
  }

  if (! (out = stream404_util_read_line (fp_backup)))
    return NULL;

  if (out[0])
    out[strlen (out) - 1] = '\0';

  if (! out[0]) {
    if (level++) {
      level = 0;
      stream404_log_error ("Unreadable or empty playlist");
      return NULL;
    }
    return playlist_builtin_backup_get_next (arg);
  }

  level = 0;

  lineno_backup++;

  stream404_log_debug ("Builtin playlist handler serving: %s", stream404_util_nullcheck (out));

  if (strncasecmp(out, "http://", 7) == 0)
    source->location = REMOTE;

  return out;
}

/* Return the current main playlist file line number */
static int
playlist_builtin_main_get_lineno (void)
{
  return lineno_main;
}

/* Return the backup main playlist file line number */
static int
playlist_builtin_backup_get_lineno (void)
{
  return lineno_backup;
}

/* Shutdown the main builtin playlist handler */
static void
playlist_builtin_main_shutdown (void)
{
  if (fp_main)
    stream404_util_fclose (fp_main);
}

/* Shutdown the backup builtin playlist handler */
static void
playlist_builtin_backup_shutdown (void)
{
  if (fp_backup)
    stream404_util_fclose (fp_backup);
}
	
/* Private function definitions */

/* Shuffle the main playlist by creating a box-unique "internal" playlist
 * and using that as the playlist */
static void
playlist_builtin_main_shuffle_playlist (void)
{
  char *newname, namespace[1024], buf[1024];
  FILE* new;

  if (! stream404_config.base_directory) {
    stream404_log_error ("Base directory is invalid");
    return;
  }

  newname = stream404_util_get_random_filename (buf, "playlist");
  snprintf (namespace, sizeof (namespace), "%s/%s",
	    stream404_config.base_directory, buf);
  new = fopen (namespace, "w+");
  if (!new) {
    stream404_log ("Error writing randomized playlist file: %s", namespace);
    return;
  }
  unlink (namespace);
  
  rand_file (fp_main, new);
  stream404_util_fclose (fp_main);

  fp_main = new;
  rewind (fp_main);
}

/* Shuffle the backup playlist by creating a box-unique "internal" playlist
 * and using that as the playlist */
static void
playlist_builtin_backup_shuffle_playlist (void)
{
  char *newname, namespace[1024], buf[1024];
  FILE* new;

  if (! stream404_config.base_directory) {
    stream404_log_error ("Base directory is invalid");
    return;
  }

  newname = stream404_util_get_random_filename (buf, "playlist");
  snprintf (namespace, sizeof (namespace), "%s/%s",
	    stream404_config.base_directory, buf);
  new = fopen (namespace, "w+");
  if (!new) {
    stream404_log ("Error writing randomized playlist file: %s", namespace);
    return;
  }
  unlink (namespace);
  
  rand_file (fp_backup, new);
  stream404_util_fclose (fp_backup);

  fp_backup = new;
  rewind (fp_backup);
}

/* Verify that the user specified main playlist actually exists */
static int
playlist_builtin_main_open_playlist (playlist_module_t* pm)
{
  struct stat st;

  if (!pm->playlist_file || !pm->playlist_file[0]) {
    stream404_log_error ("Playlist file is not set!");
    return 0;
  }

  fp_main = stream404_util_fopen_for_reading (pm->playlist_file);

  if (fp_main) {
    if (stat (pm->playlist_file, &st) == 0)
      playlist_modtime_main = st.st_mtime;

    return 1;
  } else {
    stream404_log_error ("Could not open playlist file: %s", pm->playlist_file);
    return 0;
  }
}

/* Verify that the user specified backup playlist actually exists */
static int
playlist_builtin_backup_open_playlist (playlist_module_t* pm)
{
  struct stat st;

  if (!pm->playlist_file || !pm->playlist_file[0]) {
    stream404_log_error ("Playlist file is not set!");
    return 0;
  }

  fp_backup = stream404_util_fopen_for_reading (pm->playlist_file);

  if (fp_backup) {
    if (stat (pm->playlist_file, &st) == 0)
      playlist_modtime_backup = st.st_mtime;

    return 1;
  } else {
    stream404_log_error ("Could not open playlist file: %s", pm->playlist_file);
    return 0;
  }
}

static int
playlist_builtin_main_reopen_playlist (playlist_module_t* pm)
{
  stream404_log_debug ("Reopening playlist file");

  if (fp_main) {
    stream404_util_fclose (fp_main);
    fp_main = NULL;
  }

  if (! playlist_builtin_main_open_playlist (&stream404_config.pm))
    return 0;

  /* if the file has been shortened, reset to the beginning */
  if (!playlist_builtin_line_skip (lineno_main, fp_main)) {
    if (! feof (fp_main)) {
      stream404_log_error ("Error seeking in playlist file");
      return 0;
    } else {
      stream404_log_debug ("Reached end of playlist, rewinding");
      lineno_main = 0;
      rewind (fp_main);
    }
  }

  return 1;
}

static int
playlist_builtin_backup_reopen_playlist (playlist_module_t* pm)
{
  stream404_log_debug ("Reopening playlist file");

  if (fp_backup) {
    stream404_util_fclose (fp_backup);
    fp_backup = NULL;
  }

  if (! playlist_builtin_backup_open_playlist (&stream404_config.pm))
    return 0;

  /* if the file has been shortened, reset to the beginning */
  if (!playlist_builtin_line_skip (lineno_backup, fp_backup)) {
    if (! feof (fp_backup)) {
      stream404_log_error ("Error seeking in playlist file");
      return 0;
    } else {
      stream404_log_debug ("Reached end of playlist, rewinding");
      lineno_backup = 0;
      rewind (fp_backup);
    }
  }

  return 1;
}

static int
playlist_builtin_line_skip (int lineno, FILE* fp)
{
  int i;
  char buf[1024];

  for (i = lineno; i > 0;) {
    if (! fgets (buf, 1024, fp))
      return 0;
    if (strchr (buf, '\n'))
      i--;
  }

  if (feof (fp))
    return 0;

  return 1;
}
